﻿using System;
using System.IO;
using System.IO.Compression;
using System.Diagnostics;
using System.Xml;
using System.Collections.Generic;
using System.Xml.Linq;
using System.Linq;
using System.Threading.Tasks;
using System.Collections.Concurrent;
using System.Threading;
using System.Threading.Channels;

namespace ksp_osm
{
    class Program
    {

        static IEnumerable<Node> LoadNodes(string file) =>
            ParallelOsmReader.LoadElements(file, (e, id) => e == "node")
            .Select(Node.FromXml);

        /// načte ze souboru vybrané OsmMembery a rekurzivně jejich obsah (tzn členy relací a vrcholy cest)
        static OsmMap LoadSpecializedMap(string file, params OsmMember[] members)
        {
            // na načtení jedné relace možná moc obecná funkce, ale třeba se bude hodit ještě na něco :)

            // tady si budeme držet TODO list - to co nám zatím chybí
            var todoRels = new HashSet<long>(members.Where(m => m.Type == OsmType.Relation).Select(m => m.Id));
            var todoWays = new HashSet<long>(members.Where(m => m.Type == OsmType.Way).Select(m => m.Id));
            var todoNodes = new HashSet<long>(members.Where(m => m.Type == OsmType.Node).Select(m => m.Id));

            var nodes = new Dictionary<long, Node>();
            var ways = new Dictionary<long, Way>();
            var relations = new Dictionary<long, Relation>();


            while (todoRels.Any() || todoWays.Any() || todoNodes.Any())
            {
                Console.Error.WriteLine("Reading XML");
                var elements = ParallelOsmReader.LoadElements(file, (elName, id) => {
                    switch (elName) {
                        case "node": return todoNodes.Count > 0 && todoNodes.Contains(id.Value);
                        case "way": return todoWays.Count > 0 && todoWays.Contains(id.Value);
                        case "relation": return todoRels.Count > 0 && todoRels.Contains(id.Value);
                        default: return false;
                    }
                });

                // Jestli nám chybí nějaké relace, tak je nabereme všechny najednou. I všech relací v celé Evropě je pohodlně málo.
                // Důvod je, že můžou být vnořené a pak by bylo potřeba načíst celý soubor znovu, abychom skočili jen na relace.
                var allRelations = todoRels.Any() ? new Dictionary<long, Relation>() : null;

                foreach (var e in elements)
                {
                    var id = (long)e.Attribute("id");
                    if (e.Name == "node" && todoNodes.Remove(id))
                    {
                        nodes.Add(id, Node.FromXml(e));
                    }
                    else if (e.Name == "way" && todoWays.Remove(id))
                    {
                        var way = Way.FromXml(e);
                        ways.Add(way.Id, way);
                        // přidáme si všechny vrcholy ze kterých je cesta vyrobená do TODO listu
                        foreach (var n in way.NodeRefs)
                        {
                            if (!nodes.ContainsKey(n))
                                todoNodes.Add(n);
                        }
                    }
                    else if (allRelations is object)
                    {
                        allRelations.Add(id, Relation.FromXml(e));
                    }
                }

                // vyfiltrujeme relace
                while (todoRels.Count > 0 && allRelations != null)
                {
                    var id = todoRels.First();
                    todoRels.Remove(id);
                    var r = allRelations[id];

                    // uložíme si já do výsledného seznamu
                    relations.Add(id, r);
                    // a pak si přidáme do TODO listu všechny její členy
                    foreach (var m in r.Members)
                    {
                        if (m.Type == OsmType.Node)
                        {
                            if (!nodes.ContainsKey(m.Id))
                                todoNodes.Add(m.Id);
                        }
                        else if (m.Type == OsmType.Way)
                        {
                            if (!ways.ContainsKey(m.Id))
                                todoWays.Add(m.Id);
                        }
                        else
                        {
                            if (!relations.ContainsKey(m.Id))
                                todoRels.Add(m.Id);
                        }
                    }
                }
            }

            return new OsmMap(nodes, ways, relations);
        }


        static IEnumerable<Node> NajdiPitka(string file, int relId)
        {
            var polygonMap = LoadSpecializedMap(file, new OsmMember(OsmType.Relation, relId));
            var polygon = Polygon.FromRelation(polygonMap, relId);

            return LoadNodes(file)
                   .Where(n => n.Tags.TryGetValue("drinking_water", out var drinking) && drinking == "yes" ||
                               n.Tags.TryGetValue("amenity", out var amenity) && amenity == "drinking_water")
                   .Where(n => polygon.Contains(n.Location));
        }

        static async Task<int> Main(string[] args)
        {
            if (args.Length != 2) {
                Console.Error.WriteLine("Usage: ./" + System.AppDomain.CurrentDomain.FriendlyName
                + " <file.xml.gz> <area_id>");
                return 1;
            }

            var file = args[0];
            // var relId = 434760; // brno
            // var relId = 2698607; // alpy
            var relId = int.Parse(args[1]);


            foreach (var p in NajdiPitka(file, relId)) Console.WriteLine(p);
            return 0;
        }


    }
}
